"""
This file implements two methods

run_as
Allows you to spawn a process as another user, by calling CreateProcessWithLogonW
This can be used by any regular user

run_as_with_system
Allows you to spawn a process as another user, by calling CreateProcessAsUser
This can only be used by special users such as LOCAL_SYSTEM, LOCAL_SERVICE, NETWORK_SERVICE

"""
import os
from ctypes import *
from ctypes.wintypes import *
from typing import Optional

# Microsoft constants defined in Winbase.h
_INVALID_HANDLE_VALUE = -1
CREATE_UNICODE_ENVIRONMENT = 0x00000400
LOGON32_LOGON_INTERACTIVE = 2
LOGON32_PROVIDER_WINNT50 = 3
NORMAL_PRIORITY_CLASS = 32

# Special types and pointers
CData = Array.__base__
LPBYTE = POINTER(BYTE)
LPHANDLE = POINTER(HANDLE)


class PROCESS_INFORMATION(Structure):
    """
    PROCESS_INFORMATION from https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-process_information
    This is what we get back when we create the process
    """

    __slots__ = "dwThreadId", "dwProcessId", "hThread", "hProcess"
    _fields_ = [
        (
            "hProcess",
            HANDLE,
        ),  # A handle to the newly created process. The handle is used to specify the process in all functions that perform operations on the process object.
        (
            "hThread",
            HANDLE,
        ),  # A handle to the primary thread of the newly created process. The handle is used to specify the thread in all functions that perform operations on the thread object.
        (
            "dwProcessId",
            DWORD,
        ),  # A value that can be used to identify a process. The value is valid from the time the process is created until all handles to the process are closed and the process object is freed; at this point, the identifier may be reused.
        (
            "dwThreadId",
            DWORD,
        ),  # A value that can be used to identify a thread. The value is valid from the time the thread is created until all handles to the thread are closed and the thread object is freed; at this point, the identifier may be reused.
    ]

    def __repr__(self):
        return f"ProcessInformation(dwProcessId={self.dwProcessId}, dwThreadId={self.dwThreadId}, hThread={self.hThread}, hProcess={self.hProcess})"


# Pointer to a PROCESS_INFORMATION struct
LPPROCESS_INFORMATION = POINTER(PROCESS_INFORMATION)


class STARTUPINFOW(Structure):
    """
    STARTUPINFOW struct from https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/ns-processthreadsapi-startupinfow

    """

    _fields_ = [
        ("cb", DWORD),  # The size of the structure, in bytes.
        ("lpReserved", LPWSTR),  # Reserved; must be NULL.
        (
            "lpDesktop",
            LPWSTR,
        ),  # The name of the desktop, or the name of both the desktop and window station for this process. A backslash in the string indicates that the string includes both the desktop and window station names.
        (
            "lpTitle",
            LPWSTR,
        ),  # For console processes, this is the title displayed in the title bar if a new console window is created. If NULL, the name of the executable file is used as the window title instead. This parameter must be NULL for GUI or console processes that do not create a new console window.
        (
            "dwX",
            DWORD,
        ),  # If dwFlags specifies STARTF_USEPOSITION, this member is the x offset of the upper left corner of a window if a new window is created, in pixels. Otherwise, this member is ignored.
        (
            "dwY",
            DWORD,
        ),  # If dwFlags specifies STARTF_USEPOSITION, this member is the y offset of the upper left corner of a window if a new window is created, in pixels. Otherwise, this member is ignored.
        (
            "dwXSize",
            DWORD,
        ),  # If dwFlags specifies STARTF_USESIZE, this member is the width of the window if a new window is created, in pixels. Otherwise, this member is ignored.
        (
            "dwYSize",
            DWORD,
        ),  # If dwFlags specifies STARTF_USESIZE, this member is the height of the window if a new window is created, in pixels. Otherwise, this member is ignored.
        (
            "dwXCountChars",
            DWORD,
        ),  # If dwFlags specifies STARTF_USECOUNTCHARS, if a new console window is created in a console process, this member specifies the screen buffer width, in character columns. Otherwise, this member is ignored.
        (
            "dwYCountChars",
            DWORD,
        ),  # If dwFlags specifies STARTF_USECOUNTCHARS, if a new console window is created in a console process, this member specifies the screen buffer height, in character rows. Otherwise, this member is ignored.
        (
            "dwFillAttribute",
            DWORD,
        ),  # If dwFlags specifies STARTF_USEFILLATTRIBUTE, this member is the initial text and background colors if a new console window is created in a console application. Otherwise, this member is ignored.
        (
            "dwFlags",
            DWORD,
        ),  # A bitfield that determines whether certain STARTUPINFO members are used when the process creates a window. This member can be one or more of the following values.
        (
            "wShowWindow",
            WORD,
        ),  # If dwFlags specifies STARTF_USESHOWWINDOW, this member can be any of the values that can be specified in the nCmdShow parameter for the ShowWindow function, except for SW_SHOWDEFAULT. Otherwise, this member is ignored.
        ("cbReserved2", WORD),  # Reserved for use by the C Run-time; must be zero.
        ("lpReserved2", LPBYTE),  # Reserved for use by the C Run-time; must be NULL.
        (
            "hStdInput",
            HANDLE,
        ),  # If dwFlags specifies STARTF_USESTDHANDLES, this member is the standard input handle for the process. If STARTF_USESTDHANDLES is not specified, the default for standard input is the keyboard buffer.
        (
            "hStdOutput",
            HANDLE,
        ),  # If dwFlags specifies STARTF_USESTDHANDLES, this member is the standard output handle for the process. Otherwise, this member is ignored and the default for standard output is the console window's buffer.
        (
            "hStdError",
            HANDLE,
        ),  # If dwFlags specifies STARTF_USESTDHANDLES, this member is the standard error handle for the process. Otherwise, this member is ignored and the default for standard error is the console window's buffer.
    ]


# A pointer to a STARTUPINFOW structure
LPSTARTUPINFOW = POINTER(STARTUPINFOW)


class SECURITY_ATTRIBUTES(Structure):
    # https://docs.microsoft.com/en-us/windows/win32/api/wtypesbase/ns-wtypesbase-security_attributes
    _fields_ = [
        ("nLength", DWORD),  # The size, in bytes, of this structure. Set this value to the size of the SECURITY_ATTRIBUTES structure.
        ("lpSecurityDescriptor", LPVOID),  # A pointer to a SECURITY_DESCRIPTOR structure that controls access to the object.
        ("bInheritHandle", BOOL,),  # A Boolean value that specifies whether the returned handle is inherited when a new process is created.
    ]


# A pointer to a SECURITY_ATTRIBUTES struct
LPSECURITY_ATTRIBUTES = POINTER(SECURITY_ATTRIBUTES)

# https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-createprocesswithlogonw
# Definition of the CreateProcessWithLogonW function
windll.advapi32.CreateProcessWithLogonW.restype = BOOL
windll.advapi32.CreateProcessWithLogonW.argtypes = [
    LPCWSTR,  # lpUsername
    LPCWSTR,  # lpDomain
    LPCWSTR,  # lpPassword
    DWORD,  # dwLogonFlags
    LPCWSTR,  # lpApplicationName
    LPWSTR,  # lpCommandLine (inout)
    DWORD,  # dwCreationFlags
    LPCWSTR,  # lpEnvironment  (force Unicode)
    LPCWSTR,  # lpCurrentDirectory
    LPSTARTUPINFOW,  # lpStartupInfo
    LPPROCESS_INFORMATION,  # lpProcessInfo (out)
]

# https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-createprocessasuserw
# Definitions of the CreateProcessAsUserW function
windll.advapi32.CreateProcessAsUserW.restype = BOOL
windll.advapi32.CreateProcessAsUserW.argtypes = [
    HANDLE,  # hToken
    LPCWSTR,  # lpApplicationName
    LPWSTR,  # lpCommandLine
    LPSECURITY_ATTRIBUTES,  # lpProcessAttributes
    LPSECURITY_ATTRIBUTES,  # lpThreadAttributes
    BOOL,  # bInheritHandles
    DWORD,  # dwCreationFlags
    LPVOID,  # lpEnvironment
    LPCWSTR,  # lpCurrentDirectory
    LPSTARTUPINFOW,  # # lpStartupInfo
    LPPROCESS_INFORMATION,  # lpProcessInformation
]

# https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-logonuserw
# Definitions of the LogonUserW function
windll.advapi32.LogonUserW.restype = BOOL
windll.advapi32.LogonUserW.argtypes = [
    LPCWSTR,  # lpszUsername,
    LPCWSTR,  # lpszDomain
    LPCWSTR,  # lpszPassword
    DWORD,  # dwLogonType
    DWORD,  # dwLogonProvider
    PHANDLE,  # phToken
]


def CreateProcessWithLogonW(
    lpUsername=None,
    lpDomain=None,
    lpPassword=None,
    dwLogonFlags=0,
    lpApplicationName=None,
    lpCommandLine=None,
    dwCreationFlags=0,
    lpEnvironment=None,
    lpCurrentDirectory=None,
    startupInfo=None,
):
    if lpCommandLine is not None and not isinstance(lpCommandLine, CData):
        lpCommandLine = create_unicode_buffer(lpCommandLine)
    dwCreationFlags |= CREATE_UNICODE_ENVIRONMENT
    if startupInfo is None:
        startupInfo = STARTUPINFOW(sizeof(STARTUPINFOW))
    processInformation = PROCESS_INFORMATION(_INVALID_HANDLE_VALUE, _INVALID_HANDLE_VALUE)
    success = windll.advapi32.CreateProcessWithLogonW(
        lpUsername,
        lpDomain,
        lpPassword,
        dwLogonFlags,
        lpApplicationName,
        lpCommandLine,
        dwCreationFlags,
        lpEnvironment,
        lpCurrentDirectory,
        cast(byref(startupInfo), LPSTARTUPINFOW),
        cast(byref(processInformation), LPPROCESS_INFORMATION),
    )
    if not success:
        raise WinError()
    return processInformation


def CreateProcessAsUser(username=None, domain=None, password=None, application_name=None, command_line=None, creation_flags=0, startup_info=None):
    if command_line is not None and not isinstance(command_line, CData):
        command_line = create_unicode_buffer(command_line)
    creation_flags |= CREATE_UNICODE_ENVIRONMENT
    if startup_info is None:
        startup_info = STARTUPINFOW(sizeof(STARTUPINFOW))
    process_information = PROCESS_INFORMATION(_INVALID_HANDLE_VALUE, _INVALID_HANDLE_VALUE)

    token = HANDLE()

    success = windll.advapi32.LogonUserW(username, domain, password, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_WINNT50, token)
    if not success:
        raise WinError()

    success = windll.advapi32.CreateProcessAsUserW(
        token,  # hToken
        None,  # lpApplicationName
        application_name,  # lpCommandLine
        None,  # lpProcessAttributes
        None,  # lpThreadAttributes
        True,  # bInheritHandles
        NORMAL_PRIORITY_CLASS,  # dwCreationFlags
        None,  # lpEnvironment
        None,  # lpCurrentDirectory
        cast(byref(startup_info), LPSTARTUPINFOW),
        cast(byref(process_information), LPPROCESS_INFORMATION),
    )
    if not success:
        raise WinError()
    return process_information


def run_as(username: str, password: str, application: str, domain: Optional[str] = ".") -> PROCESS_INFORMATION:
    """
    Runs a process as another user. This can only be invoked by a valid user (that can logon interactively, ie NOT LOCAL_SYSTEM, LOCAL_SERVICE, NETWORK_SERVICE, etc)

    :param username: str, username
    :param password: str, password1
    :param application: str, full path to the binary, including arguments, example: "C:\\binaries\\binary.exe --param1 1 --param2 2"
    :param domain: optional domain, . by default
    :return: PROCESS_INFORMATION
    """
    current_user = os.getlogin().upper()
    if "SYSTEM" in current_user or "SERVICE" in current_user:
        return run_as_with_system(username=username, password=password, domain=domain, application=application)
    return CreateProcessWithLogonW(username, domain, password, 0, None, application)


def run_as_with_system(username: str, password: str, application: str, domain: Optional[str] = ".") -> PROCESS_INFORMATION:
    """
    Runs a process as another user. This can only be invoked by an user with SE_INCREASE_QUOTA_NAME  privileges, like LOCAL_SYSTEM, LOCAL_SERVICE, NETWORK_SERVICE

    :param username: str, username
    :param password: str, password1
    :param application: str, full path to the binary, including arguments, example: "C:\\binaries\\binary.exe --param1 1 --param2 2"
    :param domain: optional domain, . by default
    :return: PROCESS_INFORMATION
    """
    return CreateProcessAsUser(username, domain, password, application)


def main():
    command = "C:\\windows\\system32\\notepad.exe"
    process_info = run_as("test-admin", "password", command)

    print(process_info.dwProcessId, process_info.hProcess, process_info.hThread, process_info.dwThreadId)


if __name__ == "__main__":
    main()